@@ -22,18 +22,16 @@ import sys
 import tarfile
 
 from reprounzip.common import Package, load_config
+from reprounzip import signals
 from reprounzip.unpackers.common import COMPAT_OK, COMPAT_MAYBE, \
     composite_action, target_must_exist, make_unique_name, shell_escape, \
-    select_installer, join_root, FileDownloader, get_runs
+    select_installer, join_root, FileUploader, FileDownloader, get_runs
 from reprounzip.utils import unicode_, iteritems
 
 
-def docker_escape(s):
-    return '"%s"' % (s.replace('\\', '\\\\')
-                      .replace('"', '\\"'))
-
-
 def select_image(runs):
+    """Selects a base image for the experiment, with the correct distribution.
+    """
     distribution, version = runs[0]['distribution']
     distribution = distribution.lower()
     architecture = runs[0]['architecture']
@@ -42,19 +40,19 @@ def select_image(runs):
         logging.info("wanted architecture was i686, but we'll use x86_64 with "
                      "Docker")
     elif architecture != 'x86_64':
-        logging.error("Error: unsupported architecture %s" % architecture)
+        logging.error("Error: unsupported architecture %s", architecture)
         sys.exit(1)
 
     # Ubuntu
     if distribution == 'ubuntu':
         if version != '12.04':
-            logging.warning("using Ubuntu 12.04 'Precise' instead of '%s'" %
+            logging.warning("using Ubuntu 12.04 'Precise' instead of '%s'",
                             version)
         return 'ubuntu', 'ubuntu:12.04'
 
     # Debian
     elif distribution != 'debian':
-        logging.warning("unsupported distribution %s, using Debian" %
+        logging.warning("unsupported distribution %s, using Debian",
                         distribution)
         distribution, version = 'debian', '7'
 
@@ -63,9 +61,9 @@ def select_image(runs):
     if version == '8' or version.startswith('jessie'):
         return 'debian', 'debian:jessie'
     else:
-        if version != '7' and not version.startswith('wheezy'):
-            logging.warning("using Debian 7 'Wheezy' instead of '%s'" %
-                            version)
+        if (version != '7' and not version.startswith('7.') and
+                not version.startswith('wheezy')):
+            logging.warning("using Debian 7 'Wheezy' instead of '%s'", version)
         return 'debian', 'debian:wheezy'
 
 
@@ -73,7 +71,7 @@ def write_dict(filename, dct):
     to_write = {'unpacker': 'docker'}
     to_write.update(dct)
     with filename.open('wb') as fp:
-        pickle.dump(to_write, fp, pickle.HIGHEST_PROTOCOL)
+        pickle.dump(to_write, fp, 2)
 
 
 def read_dict(filename):
@@ -92,6 +90,8 @@ def docker_setup_create(args):
         logging.critical("Target directory exists")
         sys.exit(1)
 
+    signals.pre_setup(target=target, pack=pack)
+
     # Unpacks configuration file
     tar = tarfile.open(str(pack), 'r:*')
     member = tar.getmember('METADATA/config.yml')
@@ -107,14 +107,14 @@ def docker_setup_create(args):
         base_image = args.base_image[0]
     else:
         target_distribution, base_image = select_image(runs)
-    logging.info("Using base image %s" % base_image)
-    logging.debug("Distribution: %s" % (target_distribution or "unknown"))
+    logging.info("Using base image %s", base_image)
+    logging.debug("Distribution: %s", target_distribution or "unknown")
 
     target.mkdir(parents=True)
     pack.copyfile(target / 'experiment.rpz')
 
     # Writes Dockerfile
-    logging.info("Writing %s..." % (target / 'Dockerfile'))
+    logging.info("Writing %s...", target / 'Dockerfile')
     with (target / 'Dockerfile').open('w',
                                       encoding='utf-8', newline='\n') as fp:
         fp.write('FROM %s\n\n' % base_image)
@@ -133,7 +133,7 @@ def docker_setup_create(args):
             # Installs necessary packages
             fp.write('    %s && \\\n' % installer.install_script(packages))
         logging.info("Dockerfile will install the %d software packages that "
-                     "were not packed" % len(packages))
+                     "were not packed", len(packages))
 
         # Untar
         paths = set()
@@ -152,7 +152,7 @@ def docker_setup_create(args):
                 try:
                     tar.getmember(str(datapath))
                 except KeyError:
-                    logging.info("Missing file %s" % datapath)
+                    logging.info("Missing file %s", datapath)
                 else:
                     pathlist.append(unicode_(datapath))
         tar.close()
@@ -166,6 +166,8 @@ def docker_setup_create(args):
     # Meta-data for reprounzip
     write_dict(target / '.reprounzip', {})
 
+    signals.post_setup(target=target)
+
 
 @target_must_exist
 def docker_setup_build(args):
@@ -183,11 +185,12 @@ def docker_setup_build(args):
     retcode = subprocess.call(['docker', 'build', '-t', image, '.'],
                               cwd=target.path)
     if retcode != 0:
-        logging.critical("docker build failed with code %d" % retcode)
+        logging.critical("docker build failed with code %d", retcode)
         sys.exit(1)
-    logging.info("Initial image created: %s" % image.decode('ascii'))
+    logging.info("Initial image created: %s", image.decode('ascii'))
 
     unpacked_info['initial_image'] = image
+    unpacked_info['current_image'] = image
     write_dict(target / '.reprounzip', unpacked_info)
 
 
@@ -207,18 +210,18 @@ def docker_run(args):
     # Destroy previous container
     if 'ran_container' in unpacked_info:
         container = unpacked_info.pop('ran_container')
-        logging.info("Destroying previous container %s" %
+        logging.info("Destroying previous container %s",
                      container.decode('ascii'))
         retcode = subprocess.call(['docker', 'rm', '-f', container])
         if retcode != 0:
-            logging.error("Error deleting previous container %s" %
+            logging.error("Error deleting previous container %s",
                           container.decode('ascii'))
         write_dict(target / '.reprounzip', unpacked_info)
 
-    # Use the initial image
-    if 'initial_image' in unpacked_info:
-        image = unpacked_info['initial_image']
-        logging.debug("Running from initial image %s" % image.decode('ascii'))
+    # Use the initial image directly
+    if 'current_image' in unpacked_info:
+        image = unpacked_info['current_image']
+        logging.debug("Running from image %s", image.decode('ascii'))
     else:
         logging.critical("Image doesn't exist yet, have you run setup/build?")
         sys.exit(1)
@@ -245,16 +248,101 @@ def docker_run(args):
         cmds.append(cmd)
     cmds = ' && '.join(cmds)
 
+    signals.pre_run(target=target)
+
     # Run command in container
-    logging.info("Starting container %s" % container.decode('ascii'))
-    subprocess.check_call(['docker', 'run', b'--name=' + container,
-                           '-i', '-t', image,
-                           '/bin/sh', '-c', cmds])
+    logging.info("Starting container %s", container.decode('ascii'))
+    retcode = subprocess.call(['docker', 'run', b'--name=' + container,
+                               '-i', '-t', image,
+                               '/bin/sh', '-c', cmds])
+    sys.stderr.write("\n*** Command finished, status: %d\n" % retcode)
 
     # Store container name (so we can download output files)
     unpacked_info['ran_container'] = container
     write_dict(target / '.reprounzip', unpacked_info)
 
+    signals.post_run(target=target, retcode=retcode)
+
+
+class ContainerUploader(FileUploader):
+    def __init__(self, target, input_files, files, unpacked_info):
+        self.unpacked_info = unpacked_info
+        FileUploader.__init__(self, target, input_files, files)
+
+    def prepare_upload(self, files):
+        if 'current_image' not in self.unpacked_info:
+            sys.stderr.write("Image doesn't exist yet, have you run "
+                             "setup/build?\n")
+            sys.exit(1)
+
+        self.build_directory = Path.tempdir(prefix='reprozip_build_')
+        self.docker_copy = []
+
+    def upload_file(self, local_path, input_path):
+        stem, ext = local_path.stem, local_path.ext
+        name = local_path.name
+        nb = 0
+        while (self.build_directory / name).exists():
+            nb += 1
+            name = stem + ('_%d' % nb).encode('ascii') + ext
+        name = Path(name)
+        local_path.copyfile(self.build_directory / name)
+        logging.info("Copied file %s to %s", local_path, name)
+        self.docker_copy.append((name, input_path))
+
+    def finalize(self):
+        if not self.docker_copy:
+            self.build_directory.rmtree()
+            return
+
+        from_image = self.unpacked_info['current_image']
+
+        with self.build_directory.open('w', 'Dockerfile',
+                                       encoding='utf-8',
+                                       newline='\n') as dockerfile:
+            dockerfile.write('FROM %s\n\n' % from_image.decode('ascii'))
+            for src, target in self.docker_copy:
+                # FIXME : spaces in filenames will probably break Docker
+                dockerfile.write(
+                        'COPY \\\n    %s \\\n    %s\n' % (
+                            unicode_(src),
+                            unicode_(target)))
+
+            # TODO : restore permissions?
+
+        image = make_unique_name(b'reprounzip_image_')
+        retcode = subprocess.call(['docker', 'build', '-t', image, '.'],
+                                  cwd=self.build_directory.path)
+        if retcode != 0:
+            logging.critical("docker build failed with code %d", retcode)
+            sys.exit(1)
+        else:
+            logging.info("New image created: %s", image.decode('ascii'))
+            if from_image != self.unpacked_info['initial_image']:
+                retcode = subprocess.call(['docker', 'rmi', from_image])
+                if retcode != 0:
+                    logging.warning("Can't remove previous image, docker "
+                                    "returned %d", retcode)
+            self.unpacked_info['current_image'] = image
+            write_dict(self.target / '.reprounzip', self.unpacked_info)
+
+        self.build_directory.rmtree()
+
+
+@target_must_exist
+def docker_upload(args):
+    """Replaces an input file in the container.
+    """
+    target = Path(args.target[0])
+    files = args.file
+    unpacked_info = read_dict(target / '.reprounzip')
+    input_files = unpacked_info.setdefault('input_files', {})
+
+    try:
+        ContainerUploader(target, input_files, files, unpacked_info)
+    finally:
+        write_dict(target / '.reprounzip', unpacked_info)
+
 
 class ContainerDownloader(FileDownloader):
     def __init__(self, target, files, container):
@@ -287,7 +375,7 @@ def docker_download(args):
                          "experiment?")
         sys.exit(1)
     container = unpacked_info['ran_container']
-    logging.debug("Downloading from container %s" % container.decode('ascii'))
+    logging.debug("Downloading from container %s", container.decode('ascii'))
 
     ContainerDownloader(target, files, container)
 
@@ -307,14 +395,21 @@ def docker_destroy_docker(args):
         logging.info("Destroying container...")
         retcode = subprocess.call(['docker', 'rm', '-f', container])
         if retcode != 0:
-            logging.error("Error deleting container %s" %
+            logging.error("Error deleting container %s",
                           container.decode('ascii'))
 
+    if 'current_image' in unpacked_info:
+        image = unpacked_info.pop('current_image')
+        logging.info("Destroying image %s...", image.decode('ascii'))
+        retcode = subprocess.call(['docker', 'rmi', image])
+        if retcode != 0:
+            logging.error("Error deleting image %s", image.decode('ascii'))
+
     image = unpacked_info.pop('initial_image')
-    logging.info("Destroying image...")
+    logging.info("Destroying image %s...", image.decode('ascii'))
     retcode = subprocess.call(['docker', 'rmi', image])
     if retcode != 0:
-        logging.error("Error deleting image %s" % image.decode('ascii'))
+        logging.error("Error deleting image %s", image.decode('ascii'))
 
 
 @target_must_exist
@@ -324,11 +419,15 @@ def docker_destroy_dir(args):
     target = Path(args.target[0])
     read_dict(target / '.reprounzip')
 
-    logging.info("Removing directory %s..." % target)
+    logging.info("Removing directory %s...", target)
+    signals.pre_destroy(target=target)
     target.rmtree()
+    signals.post_destroy(target=target)
 
 
 def test_has_docker(pack, **kwargs):
+    """Compatibility test: has docker (ok) or not (maybe).
+    """
     pathlist = os.environ['PATH'].split(os.pathsep) + ['.']
     pathexts = os.environ.get('PATHEXT', '').split(os.pathsep)
     for path in pathlist:
@@ -347,6 +446,8 @@ def setup(parser):
 
     setup   setup/create    creates Dockerfile (needs the pack filename)
             setup/build     builds the container from the Dockerfile
+    upload                  replaces input files in the container
+                            (without arguments, lists input files)
     run                     runs the experiment in the container
     download                gets output files from the container
                             (without arguments, lists output files)
@@ -388,7 +489,11 @@ def setup(parser):
     parser_setup.set_defaults(func=composite_action(docker_setup_create,
                                                     docker_setup_build))
 
-    # TODO : docker upload
+    # upload
+    parser_upload = subparsers.add_parser('upload', parents=[options])
+    parser_upload.add_argument('file', nargs=argparse.ZERO_OR_MORE,
+                               help="<path>:<input_file_name")
+    parser_upload.set_defaults(func=docker_upload)
 
     # run
     parser_run = subparsers.add_parser('run', parents=[options])
